#!/usr/bin/env python3
from src.background_task import BackgroundTask
from src.chat import create_chat, create_chat_with_spinner, suggest_name, invoke
from src.db import ChatDB
from src.formatting import print_message
from src.formatting import setMark
from src.highlight import SyntaxHighlighter
from src.input_reader import Chat, read_input
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML
from src.search import display_search_results
from src.spin import Spinner
from src.ui_utils import draw_horizontal_line, draw_light_horizontal_line
import html
import json
import openai
import os
import subprocess
import sys
import tiktoken
import traceback
from typing import Optional

openai.api_key = os.environ.get("OPENAI_KEY")
if not openai.api_key:
    print("Set the environment variable OPENAI_KEY to your api secret key")
    exit(1)

class App:
    def __init__(self):
        self.tasks = []
        self.chat_db = ChatDB()
        self.current_chat_id = None
        self.messages = []
        self.placeholder = ""
        self.allow_execution = False
        self.temperature = 0
        self.chats = []
        self.chat = None
        self.content = None
        self.model = "gpt-4"
        self.max_tokens = 8000

    def new_chat(self):
        self.current_chat_id = self.chat_db.create_chat("New chat")
        self.messages.clear()
        self.messages.append({
            "role": "system",
            "content": "You assist a user in a terminal emulator."})
        self.chat_db.add_message(self.current_chat_id, self.messages[0]["role"],
                self.messages[0]["content"], None, None)
        return self.current_chat_id

    def run_forever(self):
        while True:
            self.iterate()

    def delete_until_user_message(self):
         while self.messages[-1]["role"] == "assistant" or self.messages[-1]["role"] == "function" or "function_call" in self.messages[-1]:
             self.chat_db.delete_message(self.messages[-1]["id"])
             self.messages = self.messages[:-1]

    def iterate(self):
        self.temperature = 0
        self.chats = list(map(lambda x: Chat(*x), self.chat_db.list_chats()))
        setMark()
        current_chat_name = self.get_chat_name()
        user_input = self.read(current_chat_name)
        self.placeholder = ""
        if not user_input:
            # Empty query means to quit.
            return False
        if user_input.regenerate:
            # Delete the last response and preceding prompt, and act as though
            # the user retyped it.
            # prompt.
            user_input.text = self.regenerate()
        elif user_input.edit:
            # Remove the last message, set placeholder, and ask for input again.
            self.edit()
            return True
        message = user_input.text
        if user_input.query:
            new_i = self.search(user_input.query)
            if new_i is not None:
                user_input.chat_identifier = new_i
            else:
                return True
        if user_input.chat_identifier:
            if user_input.chat_identifier < 0:
                print_formatted_text(HTML(f'<b>Starting a new chat</b>'))
                self.current_chat_id = self.new_chat()
            else:
                self.switch(user_input.chat_identifier)
            if not message:
                return True
        elif self.current_chat_id is None:
            self.current_chat_id = self.new_chat()
        if not message:
            return False
        if len(self.messages) == 1:
            self.assign_name(message)

        self.send_message(message)
        return True

    def send_message(self, message):
        draw_light_horizontal_line()
        setMark()

        self.messages.append(
            {"role": "user", "content": message},
        )
        message_id = self.chat_db.add_message(self.current_chat_id, "user", message, None, None)
        self.messages[-1]["id"] = message_id
        sanitized = [
                {k: v for k, v in message.items() if k != 'id'} 
                for message in self.messages]
        if self.allow_execution:
            execute_command.sloppy = False
            create_file.sloppy = False
            execute_python.sloppy = True
            functions = [execute_command, create_file, execute_python]
        else:
            functions = []
        self.chat = create_chat_with_spinner(sanitized, self.temperature, functions, self.model)
        self.content = ""
        sh = SyntaxHighlighter()
        fspinner = None
        try:
            # There can be more than one chat when there's a function call.
            while self.read_response(functions, sh):
                pass
            sh.eof()
            print("")
            print("")
        except KeyboardInterrupt:
            sh.eof()
            self.chat.close()
            print("")
        self.commit_ordinary()

        # Check if any tasks are done.
        self.check_tasks()

    def check_tasks(self):
        for task in self.tasks:
            if task.done():
                (chat_id, name) = task.result()
                self.chat_db.set_chat_name(chat_id, name)
            self.tasks = [task for task in self.tasks if not task.done()]

    def get_chat_name(self):
        if self.current_chat_id is None:
            return "New chat"
        else:
            return self.chat_db.get_chat_name(self.current_chat_id)

    def read(self, current_chat_name):
        try:
            user_input = read_input(
                    self.chats,
                    current_chat_name,
                    len(self.messages) > 1,
                    self.placeholder,
                    self.allow_execution,
                    self.usage(),
                    self.max_tokens)
            self.allow_execution = user_input.allow_execution
            return user_input 
        except KeyboardInterrupt:
            print("^C")
            exit(0)

    def message_content(self, message):
        c = message["content"]
        if c is not None:
            return c
        f = message["function_call"]
        if f is not None:
            return json.dumps(f)
        return 0

    def usage(self):
        text = ""
        contents = [self.message_content(message) for message in self.messages]
        text = "\n".join(contents)
        encoding = tiktoken.encoding_for_model(self.model)
        return len(encoding.encode(text))

    def regenerate(self):
        print("Regenerating response...")
        self.delete_until_user_message()
        text = self.messages[-1]["content"]
        self.temperature = 0.5
        self.chat_db.delete_message(self.messages[-1]["id"])
        self.messages = self.messages[:-1]
        return text

    def edit(self):
        print("Editing previous message...")
        self.delete_until_user_message()
        self.placeholder = self.messages[-1]["content"]
        self.chat_db.delete_message(self.messages[-1]["id"])
        self.messages = self.messages[:-1]

    def search(self, query):
        """Returns id of chat to switch to or else None.
        A negative id means to create a new chat."""
        i = None
        cursor = None
        PAGE_SIZE = 10
        while True:
            search_results, cursor = self.chat_db.search_messages(query, cursor, PAGE_SIZE)
            if not len(search_results):
                print("No results")
                return None
            i = display_search_results(query, search_results, self.chat_db)
            if i is None:
                cursor += PAGE_SIZE
                continue
            if i < 0:
                return -1
            return i

    def switch(self, chat_id):
        """Replace self.messages with contents of chat_id and also print the
        messages to the screen. As a side-effect, set self.current_chat_id."""
        self.current_chat_id = chat_id
        n = self.chat_db.num_messages(self.current_chat_id)
        self.messages = []
        for i in range(1, n):
            (role, self.content, time, message_id, deleted, fname, fargs) = self.chat_db.get_message_by_index(self.current_chat_id, i)
            if self.content is not None:
                m = {
                    "id": message_id,
                    "role": role,
                    "content": self.content
                    }
                if role == "function":
                    m["name"] = fname
                else:
                    if role == "user":
                        draw_horizontal_line()
                    else:
                        draw_light_horizontal_line()
                    print_message(time, role, self.content.rstrip(), deleted)
                    print("")
                self.messages.append(m)
            elif fname and fargs:
                self.messages.append({
                    "id": message_id,
                    "role": "assistant",
                    "content": None,
                    "function_call": {
                        "name": fname,
                        "arguments": fargs}})

    def assign_name(self, message):
        # The chat needs a name
        self.tasks.append(BackgroundTask(lambda: suggest_name(self.current_chat_id, message)))

    def handle_function_call(self, functions, fspinner, call_name, call_args):
        try:
            if fspinner:
                fspinner.stop()
            return (None, invoke(functions, call_name, call_args), None)
        except Exception as e:
            return (None, None, str(e))

    def stop_unexpectedly(self, finish_reason):
        print("")
        print(f"Stopping because {finish_reason}")

    def read_response(self, functions, sh):
        """Read an entire response and handle it. Return True to call this again."""
        fspinner = None
        try:
            call_name = None
            call_args = None
            function_output = None
            error_output = None
            for resp in self.chat:
                if resp.choices[0].finish_reason:
                    finish_reason = resp.choices[0].finish_reason
                    if finish_reason == "function_call" and call_name and call_args:
                        fspinner, function_output, error_output = self.handle_function_call(
                                functions,
                                fspinner,
                                call_name,
                                call_args)
                    elif finish_reason != "stop":
                        self.stop_unexpectedly(finish_reason)
                    break
                elif "content" in resp.choices[0].delta and resp.choices[0].delta.content:
                    self.content = self.handle_content(resp, self.content, sh)
                elif "function_call" in resp.choices[0].delta and resp.choices[0].delta.function_call and self.allow_execution:
                    call_name, call_args, fspinner = self.accrue_function_call(
                            resp, call_name, call_args, fspinner)


            return self.commit_special(
                    call_name,
                    call_args,
                    function_output,
                    error_output,
                    functions)
        finally:
            if fspinner:
                fspinner.stop()
                fspinner = None

    def handle_content(self, resp, content, sh):
        chunk = resp.choices[0].delta.content
        content += chunk
        sh.put(chunk)
        return content

    def accrue_function_call(self, resp, call_name, call_args, fspinner):
        if "name" in resp.choices[0].delta.function_call:
            if call_name:
                call_name += resp.choices[0].delta.function_call.name
            else:
                fspinner = Spinner()
                call_name = resp.choices[0].delta.function_call.name
        if "arguments" in resp.choices[0].delta.function_call:
            if call_args:
                call_args += resp.choices[0].delta.function_call.arguments
            else:
                call_args = resp.choices[0].delta.function_call.arguments
        return (call_name, call_args, fspinner)

    def commit_special(self, call_name, call_args, function_output,
            error_output, functions):
        if call_name and call_args:
            self.commit_function_call_request(call_name, call_args);
        if function_output:
            self.commit_function_output(functions, call_name, function_output)
            return True
        if error_output:
            return self.commit_error(error_output, functions)
        return False

    def commit_ordinary(self):
        if self.content is not None:
            self.messages.append({"role": "assistant", "content": self.content})
        message_id = self.chat_db.add_message(
                self.current_chat_id,
                self.messages[-1]["role"],
                self.messages[-1]["content"],
                None,
                None)
        self.messages[-1]["id"] = message_id

    def commit_function_call_request(self, call_name, call_args):
        # Record that a function call was requested
        self.messages.append({
            "role": "assistant",
            "content": None,
            "function_call": {
                "name": call_name,
                "arguments": call_args}})
        message_id = self.chat_db.add_message(self.current_chat_id,
                self.messages[-1]["role"], self.messages[-1]["content"], call_name, call_args)
        self.messages[-1]["id"] = message_id

    def commit_function_output(self, functions, call_name, function_output):
        # Record the output of the function call
        self.messages.append({
            "role": "function",
            "name": call_name,
            "content": function_output})
        message_id = self.chat_db.add_message(
                self.current_chat_id,
                self.messages[-1]["role"],
                self.messages[-1]["content"],
                call_name,
                None)
        self.messages[-1]["id"] = message_id
        sanitized = [
                {k: v for k, v in message.items() if k != 'id'} 
                for message in self.messages]
        self.chat = create_chat(sanitized, self.temperature, functions, self.model)

    def commit_error(self, error_output, functions):
        # Something went wrong
        print(f'Error: {error_output}')
        auto_retry = False
        self.messages.append({
            "role": "system",
            "content": error_output})
        message_id = self.chat_db.add_message(
                self.current_chat_id,
                self.messages[-1]["role"],
                self.messages[-1]["content"],
                None,
                None)
        self.messages[-1]["id"] = message_id
        if auto_retry:
            sanitized = [
                    {k: v for k, v in message.items() if k != 'id'} 
                    for message in self.messages]
            print(sanitized)
            self.chat = create_chat(sanitized, self.temperature, functions, self.model)
            return True
        else:
            print(f"An error was encountered while executing a function: {error_output}")
            self.content = None
            return False

def main():
    print("Welcome to gptline! Enter a question and press option-Enter to send it.")
    app = App()
    app.run_forever()

def execute_command(command_line: str, input_string: str):
    """
    Executes a unix command at the shell. Note that it does not run in a TTY so interactive commands will not work.

    Args:
        command_line: Unix command to execute. For example, "ls -l". This string will be passed to system().
        input_string: A string containing input to send to stdin.
    """
    print("")
    print("ChatGPT wants to run the following command:")
    print(command_line)
    print("")
    ok = input(f"OK to run (y/n)? ")
    if ok != "y":
        print("Not running it")
        return "Error: user denied permission to execute function call"
    print("Executing...")

    try:
        process = subprocess.Popen(command_line, shell=True, cwd="/tmp", stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
        output, _ = process.communicate(input=input_string)
        print("Program output:")
        print(output)
        print("")
        return output.strip()
    except Exception as e:
        print(f'Exception while executing provided code: {e}')
        return str(e)

def create_file(name: str, content: str):
    """
    Creates a file with the given name and contents.

    Args:
        name: A file name. This will be relative to a private temporary directory.
        content: A string to write to the file.
    """
    print(f"ChatGPT wants to create a file called {name} with the following contents:")
    sh = SyntaxHighlighter()
    sh.put("```")
    sh.put(content)
    sh.put("")
    sh.put("```")
    sh.eof()
    ok = input(f"OK to create (y/n)? ")
    if ok != "y":
        print("Not running it")
        return "Error: user denied permission to create file"
    file_path = os.path.join('/tmp', name)

    try:
        # Open the file in write mode and write the content
        with open(file_path, 'w') as file:
            file.write(content)
        return f"File {name} created successfully."
    except Exception as e:
        return f"Error creating file: {str(e)}"

def execute_python(code: str, input_string: Optional[str]):
    """
    Executes python code. Outputs values from stdout. If you need a computed value make sure to print() it.

    Args:
        code: Python code to execute.
        input_string: A string containing input to send to stdin.
    """
    print("ChatGPT wants to run the following Python program:")
    sh = SyntaxHighlighter()
    sh.put("```python\n")
    sh.put(code)
    if not code.endswith("\n"):
        sh.put("\n")
    sh.put("```")
    sh.eof()
    print("")
    if input_string:
        print("With this supplied as input:")
        print(input_string)
        print("")
    ok = input(f"OK to run (y/n)? ")
    if ok != "y":
        print("Not running it")
        return "Error: user denied permission to execute function call"
    print("Executing...")
    try:
        process = subprocess.Popen(["python3", "-c", code], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        stdout, stderr = process.communicate(input=input_string if input_string else "")
        print("Program output:")
        print(stdout + stderr)
        if stderr:
            return f"Error: {stderr}"
        return stdout
    except Exception as e:
        return f"Error: {str(e)}"



from blessings import Terminal
from dataclasses import dataclass
import re
import sys
import termios
import tty
import uuid
from pygments.lexers import get_lexer_by_name
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.style import Style
from pygments.token import Token
from pygments.styles import get_style_by_name


@dataclass
class Theme:
    code_block_banner_bg: str
    code_block_banner_fg: str
    code_block_bg: str
    inline_code: str

def foreground_color(r, g, b):
    return f"\033[38;2;{r};{g};{b}m"

def background_color(r, g, b):
    return f"\033[48;2;{r};{g};{b}m"

class SyntaxHighlighter:
    def __init__(self):
        self.term = Terminal()
        self.buffer = ""
        self.is_bold = False
        self.is_italic = False
        self.is_underline = False
        self.is_inline_code = False
        self.is_block_code = False
        self.block_id = None
        self.will_enter_block_code = False
        self.prev_attrs = []
        self.markup_eligible = True
        self.line = ""
        self.language = None
        darkTheme = Theme(background_color(55, 55, 55),
                          foreground_color(215, 215, 215),
                          background_color(15, 15, 15),
                          self.term.bold)

        lightTheme = Theme(background_color(215, 215, 215),
                           foreground_color(15, 15, 15),
                           background_color(240, 240, 240),
                           self.term.bold)

        if is_background_dark():
            self.theme = darkTheme
            self.formatter = TerminalTrueColorFormatter()
        else:
            self.theme = lightTheme
            self.formatter = TerminalTrueColorFormatter()
        self.lexer = None
        self.lex_buffer = ""

    def put(self, s):
        self._put(s)
        sys.stdout.flush()

    def _put(self, s):
        if len(s) == 0:
            return

        self.buffer += s
        while len(self.buffer) > 0:
            if not self.is_block_code and self.buffer.startswith("**"):
                self.is_bold = not self.is_bold
                self.buffer = self.buffer[2:]
            elif not self.is_block_code and self.buffer.startswith("*") and len(self.buffer) > 1 and self.buffer[1] != "*" and self.buffer[1] != " " and self.buffer[1] != "\n":
                self.is_italic = not self.is_italic
                self.buffer = self.buffer[1:]
            elif not self.is_block_code and self.buffer == "*":
                return
            elif not self.is_block_code and self.buffer.startswith("_") and self.markup_eligible:
                self.is_underline = not self.is_underline
                self.buffer = self.buffer[1:]
            elif not self.is_inline_code and self.buffer == "`":
                # A single backtick is ambiguous. Could be an inline block or a code block.
                return
            elif not self.is_block_code and not self.is_inline_code and self.buffer.startswith("`") and len(self.buffer) > 1 and self.buffer[1] != "`" and self.line and self.line[-1] == ' ':
                # Start an inline block
                self.is_inline_code = True
                self.buffer = self.buffer[1:]
            elif not self.is_block_code and self.is_inline_code and self.buffer.startswith("\n"):
                # End an inline block at newline to avoid it running away
                self.is_inline_code = False
            elif not self.is_block_code and self.is_inline_code and self.buffer.startswith("`"):
                # End an inline block
                self.is_inline_code = False
                self.buffer = self.buffer[1:]
            elif not self.is_block_code and not self.will_enter_block_code and re.match(r'^[ ]{0,8}$', self.line) is not None and self.buffer.startswith("```"):
                # Start a code block
                if '\n' not in self.buffer[1:]:
                    # Ignore language, which follows ```
                    return
                self.is_block_code = True
                self.block_id = uuid.uuid4()
                self.will_enter_block_code = True
                # Remove up to and including first newline
                first, second = self.buffer.split('\n', 1)
                if len(first) > 4:
                  # Print language
                  print(self.theme.code_block_banner_bg + self.theme.code_block_banner_fg, end='')
                  print(first[3:] + self.erase_line() + "    " + self.copyButton(self.block_id) + " " + self.hyperlink(f"iterm2:copy-block?block={self.block_id}", "Copy to clipboard") + self.term.normal)
                  self.language = first[3:]
                else:
                  self.language = "text/plain"
                try:
                    self.lexer = get_lexer_by_name(first[3:])
                except:
                    pass
                self.prev_attrs = None
                self.buffer = second
            elif self.is_block_code and not self.will_enter_block_code and self.buffer.startswith("```\n"):
                # End a code block
                if not self.will_enter_block_code:
                    self.emit(self.end_block(self.block_id) + self.term.normal)
                self.block_id = None
                self.will_enter_block_code = False
                self.buffer = self.buffer[4:]
                self.is_block_code = False
                self.lexer = None
            elif self.is_block_code and (self.buffer == "`" or self.buffer == "``"):
                # Too soon to tell
                return
            elif not self.is_block_code and re.match(r"\[.*\]\(.*\)", self.buffer):
                link_text, link_url = re.findall(r"\[(.*?)\]\((.*?)\)", self.buffer)[0]
                self.buffer = self.buffer[len(link_text) + len(link_url) + 4:]
                self.emit(self.hyperlink(link_url, link_text))
            elif not self.is_block_code and (self.buffer.startswith('[') or self.buffer == "``") and '\n' not in self.buffer:
                return
            else:
                self.printFirst()

    def copyButton(self, block_id):
        return self.osc(1337) + f'Button=type=copy;block={block_id}' + self.st()

    def erase_line(self):
        return '\033[K'

    def osc(self, n):
        return f'\033]{n};'

    def st(self):
        return f'\a'

    def hyperlink(self, href, anchor):
        return self.osc(8) + f';{href}' + self.st() + anchor + self.osc(8) + ';' + self.st()

    def start_block(self, id, type):
        return self.osc(1337) + f'Block=attr=start;id={id};type={type}' + self.st()

    def end_block(self, id):
        return self.osc(1337) + f'Block=attr=end;id={id};render=0' + self.st()

    def eof(self):
        while len(self.buffer) > 0:
            self.printFirst()
        if self.is_block_code:
            self.emit(self.end_block(self.block_id))

    def lex(self):
        highlighted_code = highlight(self.lex_buffer, self.lexer, self.formatter)
        highlighted_code = highlighted_code.replace("\n", self.theme.code_block_bg + self.erase_line() + "\n")
        print(highlighted_code, end='')
        self.lex_buffer = ""

    def printFirst(self):
        attrs = []
        if self.is_block_code:
            attrs.append(self.theme.code_block_bg)
        if self.will_enter_block_code:
            self.emit(self.theme.code_block_bg + self.erase_line())
            self.emit(self.start_block(self.block_id, self.language))
            self.will_enter_block_code = False
            attrs.append(self.theme.code_block_bg)
        if self.lexer:
            if self.buffer[0] == "\n":
                self.lex()
                self.line = ""
            else:
                self.lex_buffer += self.buffer[0]
            self.buffer = self.buffer[1:]
            return

        if self.is_bold:
            attrs.append(self.term.bold)
        elif self.is_italic:
            attrs.append(self.term.italic)
        elif self.is_underline:
            attrs.append(self.term.underline)
        elif self.is_inline_code:
            attrs.append(self.theme.inline_code)
        if self.prev_attrs != attrs:
            self.emit(self.term.normal + "".join(attrs) + self.buffer[0])
            self.prev_attrs = attrs
        else:
            self.emit(self.buffer[0])
        self.markup_eligible = self.buffer[0] == ' ' or self.buffer[0] == '\n'

        if self.buffer[0] == "\n":
            self.line = ""
            self.emit(self.erase_line())
        else:
            self.line += self.buffer[0]
        self.buffer = self.buffer[1:]

    def emit(self, chunk):
        print(chunk, end='')

def is_background_dark():
    # Save the current terminal settings
    old_settings = termios.tcgetattr(sys.stdin)

    try:
        # Set the terminal to raw mode
        tty.setraw(sys.stdin)

        # Write the query control sequence to the terminal
        sys.stdout.write('\033]11;?\033\\')
        sys.stdout.flush()

        # Read the response from the terminal
        response = ''
        while True:
            char = sys.stdin.read(1)
            response += char
            if response.endswith('\033\\'):
                break

        # Extract the background color information from the response
        match = re.search(r'rgb:([0-9a-fA-F]{4})/([0-9a-fA-F]{4})/([0-9a-fA-F]{4})', response)
        if match:
            r = int(match.group(1), 16)
            g = int(match.group(2), 16)
            b = int(match.group(3), 16)
            luminance = (0.299 * r + 0.587 * g + 0.114 * b) / 1000
            return luminance < 0.5

        return False

    finally:
        # Restore the terminal settings
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)



from dataclasses import dataclass
from prompt_toolkit import PromptSession, print_formatted_text, Application
from prompt_toolkit.application.current import get_app
from prompt_toolkit.formatted_text import HTML, FormattedText
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.widgets import Frame
from src.formatting import formattedTime
from src.ui_utils import draw_horizontal_line
from typing import Optional
import html

@dataclass
class UserInput:
    text: Optional[str] = None
    # -1 to create a new chat.
    chat_identifier: Optional[int] = None
    query: Optional[str] = None
    regenerate = False
    edit = False
    allow_execution = False

@dataclass
class Chat:
    chat_identifier: int
    name: str
    last_update: str
    num_messages: int


# Returns UserInput
def read_input(chats, current_chat_name, can_regen, placeholder, allow_execution, used, max_tokens):
    result = UserInput()
    result.allow_execution = allow_execution

    session = PromptSession()
    kb = KeyBindings()

    # Add a binding for the 'Enter' key to insert a newline character
    @kb.add(Keys.Enter)
    def _(event):
        event.current_buffer.insert_text('\n')

    # Add a binding for ^D to accept the input.
    @kb.add(Keys.ControlD)
    def _(event):
        buffer = event.current_buffer
        if buffer.cursor_position < len(buffer.text):
            buffer.delete()
        else:
            buffer.validate_and_handle()

    SHOW_CHAT_LIST = "$$$SHOW_CHAT_LIST"
    NEW_CHAT = "$$$NEW_CHAT"
    SEARCH = "$$$SEARCH"
    REGENERATE = "$$$REGENERATE"
    EDIT = "$$$EDIT"
    TOGGLE_SETTING = "$$$TOGGLE_SETTING"

    @kb.add(Keys.F2)
    def _(event):
        app = get_app()
        app.exit(result=SHOW_CHAT_LIST)

    @kb.add(Keys.F3)
    def _(event):
        app = get_app()
        app.exit(result=NEW_CHAT)

    @kb.add(Keys.F4)
    def _(event):
        app = get_app()
        app.exit(result=SEARCH)

    if can_regen:
        @kb.add(Keys.F5)
        def _(event):
            app = get_app()
            app.exit(result=REGENERATE)
        @kb.add(Keys.F6)
        def _(event):
            app = get_app()
            app.exit(result=EDIT)

    @kb.add(Keys.F7)
    def _(event):
        result.allow_execution = not result.allow_execution
        if result.allow_execution:
            print("Command execution enabled.")
        else:
            print("Command execution disabled.")
        app = get_app()
        app.exit(result=TOGGLE_SETTING)
    def read_search_query():
        search_session = PromptSession()
        return search_session.prompt(HTML("<b>Search: </b>"))

    def pick_chat():
        print_formatted_text(HTML(f'<b>Select a chat:</b>'))
        show_list = True
        base = 0
        PAGE_SIZE = Application().output.get_size().rows - 2
        while True:
            i = 0
            for chat in chats[base:]:
                i += 1
                if i == PAGE_SIZE + 1:
                    break
                if show_list:
                    time = formattedTime(chat.last_update)
                    n = chat.num_messages - 1
                    s = "s" if n > 1 else ""
                    print_formatted_text(HTML(f'{i}: <b>{html.escape(chat.name)}</b> ({time}, {n} message{s})'))
            show_list = True

            kb = KeyBindings()
            user_pressed_esc = [False]
            @kb.add(Keys.ControlD)
            def _(event):
                user_pressed_esc[0] = True
                event.app.exit()

            # Prompt the user for their selection
            picker_session = PromptSession(key_bindings=kb)
            selection = picker_session.prompt(HTML("<b>Enter chat number, press 'return' for more, + to create a new chat, or '^D' to cancel: </b>"))
            if user_pressed_esc[0]:
                return

            # Handle the user's selection
            if selection.isdigit():
                index = int(selection) + base - 1
                result.chat_identifier = chats[index].chat_identifier
                print_formatted_text(HTML(f'<b>Switched to: {html.escape(chats[index].name)}</b>'))
                return
            elif selection == '':
                if len(chats) > base + PAGE_SIZE:
                    base += PAGE_SIZE
                else:
                    print_formatted_text(HTML('<b>No more chats.</b>'))
                    show_list = False
            elif selection == '+':
                result.chat_identifier = -1
                return
            else:
                print_formatted_text(HTML('<b>Selection canceled or invalid choice.</b>'))


    # Create a frame around the default buffer
    frame = Frame(body=Window(content=BufferControl(buffer=session.default_buffer)))

    # Set the layout of the session to use the frame
    session.layout = Layout(container=frame)

    # Prompt the user for input
    draw_horizontal_line()
    try:
        while True:
            def bottom_toolbar():
                text = f'[{html.escape(current_chat_name)}] <b>M-⏎</b>: Send  <b>F2</b>: Switch chat  <b>F3</b>: New chat  <b>F4</b>: Search'
                if can_regen:
                    text += "  <b>F5</b>: Regenerate"
                    text += "  <b>F6</b>: Edit Last"
                if result.allow_execution:
                    text += "  <b>F7</b>: Disable Execution"
                else:
                    text += "  <b>F7</b>: Enable Execution"
                text += f'  {used}/{max_tokens} tokens'
                return HTML(text)
            value = session.prompt(
                    "",
                key_bindings=kb,
                multiline=True,
                bottom_toolbar=bottom_toolbar,
                default=placeholder
            )
            if value == SHOW_CHAT_LIST:
                pick_chat()
                if result.chat_identifier is not None:
                    return result
            elif value == NEW_CHAT:
                result.chat_identifier = -1
                return result
            elif value == REGENERATE:
                result.regenerate = True
                return result
            elif value == TOGGLE_SETTING:
              continue
            elif value == EDIT:
                result.edit = True
                return result
            elif value == SEARCH:
                try:
                    result.query = read_search_query()
                    if result.query is not None:
                        return result
                except EOFError:
                    pass
            else:
                result.text = value
                return result
    except Exception as e:
        print(e)
        return None



import sqlite3
from typing import Optional
import fcntl
import os
import sys

def fullpath(file):
    xdg_root = os.getenv("XDG_ROOT")
    if xdg_root:
        return os.path.join(xdg_root, file)
    else:
        home_dir = os.path.expanduser("~")
        return os.path.join(home_dir, file)

lock_fd = None

class ChatDB:
    def __init__(self, db_file=".chatgpt.db"):
        db_path = fullpath(db_file)
        self.conn = sqlite3.connect(db_path)
        self.conn.execute("PRAGMA journal_mode=WAL")  # Enable WAL mode
        self.cursor = self.conn.cursor()
        self.create_schema()

    def create_schema(self):
        query = """
        CREATE TABLE IF NOT EXISTS chats (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT,
            last_update TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        )
        """
        self.conn.execute(query)

        query = """
        CREATE TABLE IF NOT EXISTS messages (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            chat_id INTEGER,
            role TEXT,
            content TEXT NULL,
            time TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            deleted INTEGER DEFAULT 0,
            function_call_name TEXT,
            function_call_arguments TEXT,
            FOREIGN KEY (chat_id) REFERENCES chats (id)
        )
        """
        self.conn.execute(query)

        query = """
        CREATE VIRTUAL TABLE IF NOT EXISTS messages_fts USING FTS5 (
            message_id UNINDEXED,
            content,
            content_rowid,
        )
        """
        self.conn.execute(query)

    def create_chat(self, name=None):
        query = "INSERT INTO chats (name) VALUES (?)"
        self.cursor.execute(query, (name,))
        chat_id = self.cursor.lastrowid  # Access lastrowid from the cursor
        self.conn.commit()
        return chat_id

    def add_message(self, chat_id: int, role: str, content: str, function_call_name: Optional[str], function_call_arguments: Optional[str]):
        query = "INSERT INTO messages (chat_id, role, content, function_call_name, function_call_arguments) VALUES (?, ?, ?, ?, ?)"
        with self.conn:
            self.cursor.execute(query, (chat_id, role, content, function_call_name, function_call_arguments))
            last_message_id = self.cursor.lastrowid

            if content is not None and role != "function":
                fts_query = "INSERT INTO messages_fts (message_id, content) VALUES (?, ?)"
                self.conn.execute(fts_query, (last_message_id, content.lower()))

            query = f"UPDATE chats SET last_update = CURRENT_TIMESTAMP WHERE id = ?"
            self.cursor.execute(query, (chat_id, ))

        return last_message_id

    def num_messages(self, chat_id: int) -> int:
        query = "SELECT COUNT(*) FROM messages WHERE chat_id = ?"
        result = self.conn.execute(query, (chat_id,)).fetchone()
        return result[0]

    def get_message_by_id(self, message_id: int):
        query = "SELECT role, content, time, id, deleted FROM messages WHERE id = ?"
        result = self.conn.execute(query, (message_id,)).fetchone()
        if result:
            return result
        else:
            raise IndexError("Index out of range")

    def get_message_by_index(self, chat_id: int, index: int):
        query = "SELECT role, content, time, id, deleted, function_call_name, function_call_arguments FROM messages WHERE chat_id = ? ORDER BY id LIMIT 1 OFFSET ?"
        result = self.conn.execute(query, (chat_id, index)).fetchone()
        if result:
            return result
        else:
            raise IndexError("Index out of range")

    def list_chats(self):
      query = """
      SELECT chats.id, chats.name, chats.last_update, COUNT(messages.id) as num_messages
      FROM chats
      LEFT JOIN messages ON chats.id = messages.chat_id
      GROUP BY chats.id
      HAVING COUNT(messages.id) > 1
      ORDER BY chats.id DESC
      """
      result = self.conn.execute(query).fetchall()
      return result

    def set_chat_name(self, chat_id: int, name: str):
        query = "UPDATE chats SET name = ? WHERE id = ?"
        self.conn.execute(query, (name, chat_id))
        self.conn.commit()

    def get_chat_name(self, chat_id):
        query = "SELECT name FROM chats WHERE id = ?"
        cursor = self.conn.execute(query, (chat_id,))
        result = cursor.fetchone()
        if result:
            return result[0]
        else:
            return None

    def delete_message(self, message_id: int):
        query = """
        UPDATE messages
        SET deleted = 1
        WHERE id = ?
        """
        self.conn.execute(query, (message_id,))
        self.conn.commit()

    def search_messages(self, query: str, pagination_token: int, limit: int):
        fts_query = """
            SELECT m.id, m.chat_id, snippet(messages_fts, 1, '\ue000', '\ue001', '...', 16) AS snippet 
            FROM messages_fts 
            JOIN messages m ON messages_fts.message_id = m.id 
            WHERE messages_fts.content MATCH ?
            LIMIT ?
            OFFSET ?
        """

        if pagination_token is None:
            offset = 0
        else:
            offset = pagination_token

        parameters = (query.lower(), limit, offset)

        result = self.conn.execute(fts_query, parameters)

        message_ids = []
        chat_ids = []
        snippets = {}
        for row in result:
            message_ids.append(int(row[0]))
            snippets[int(row[0])] = row[2]
            chat_ids.append(int(row[1]))

        # Sort chat IDs
        sorted_chat_ids = sorted(set(chat_ids))

        messages_by_chat = []
        for chat_id in sorted_chat_ids:
            message_ids_for_chat = [(message_ids[i], snippets[message_ids[i]]) for i in range(len(message_ids)) if chat_ids[i] == chat_id]
            messages_by_chat.append((chat_id, message_ids_for_chat))

        # Determine the pagination token
        next_offset = offset + limit
        has_more_results = len(message_ids) > next_offset
        pagination_token = next_offset if has_more_results else None

        return messages_by_chat, offset + len(message_ids)




from annoy import AnnoyIndex
from src.db import Embedding, fullpath
import openai
import os
import pickle
import queue
import re
import sqlite3
import threading

class VectorDatabase:
    def __init__(self, db_file=".chatgpt-vectors.db"):
        db_path = fullpath(db_file)
        self.conn = sqlite3.connect(db_path)
        self.conn.execute("PRAGMA journal_mode=WAL")  # Enable WAL mode
        self.cursor = self.conn.cursor()
        self.create_schema()
        self.index = AnnoyIndex(1536, 'angular')
        self.embeddings = self.get_embeddings()
        for i, embedding in enumerate(self.embeddings):
            self.index.add_item(i, embedding.vector)
        self.index.build(10)

    def create_schema(self):
        query = """
        CREATE TABLE IF NOT EXISTS embeddings (
            id INTEGER PRIMARY KEY,
            chat_id INTEGER,
            message_id INTEGER,
            vector BLOB,
            phrase TEXT
        );
        """
        self.conn.execute(query)
        query = """
        CREATE TABLE IF NOT EXISTS progress (
            id INTEGER PRIMARY KEY,
            last_message_id INTEGER
        );
        """
        self.conn.execute(query)

    def last_message_id(self):
        query = """
        SELECT last_message_id from progress
        """
        rows = self.conn.execute(query).fetchall()
        if not rows:
            return -1
        return rows[0][0]

    def commit_message(self, message_id):
        # Upsert into progress table
        query = """
        INSERT OR REPLACE INTO progress (id, last_message_id) VALUES (1, ?)
        """
        self.conn.execute(query, (message_id,))

        self.conn.commit()

    def add_embedding(self, vector, chat_id, message_id, phrase):
        query = """
        INSERT INTO embeddings (chat_id, message_id, vector, phrase) VALUES (?, ?, ?, ?)
        """
        # Serialize the vector to a binary format using pickle
        serialized_vector = pickle.dumps(vector)
        self.conn.execute(query, (chat_id, message_id, serialized_vector, phrase))
        self.conn.commit()

    def get_embeddings(self):  # Return List<Embedding>
        query = """
        SELECT * FROM embeddings
        """
        rows = self.conn.execute(query).fetchall()
        # Deserialize the vectors and return a list of Embedding objects
        return [Embedding(pickle.loads(row[3]), row[1], row[2]) for row in rows]

    def add(self, embedding: Embedding, phrase):  # Save to db
        self.add_embedding(
            embedding.vector, embedding.chat_id, embedding.message_id, phrase)
        self.embeddings.append(embedding)
        self.index.add_item(len(self.embeddings) - 1, embedding.vector)

    def find_neighbors(self, embedding: list, max_distance, max_results):
        # Annoy uses cosine similarity, so the distance is in the range [0, 2],
        # where 0 means identical and 2 means completely dissimilar
        ids, distances = self.index.get_nns_by_vector(
            embedding, max_results, include_distances=True)
        return [self.embeddings[i]
            for (i, distance) in zip(ids, distances)
            if distance <= max_distance]

def compute_embedding(text_string):
    model_id = "text-embedding-ada-002"
    return openai.Embedding.create(input=text_string, model=model_id)['data'][0]['embedding']

def chunk_string(text_string):
    if not text_string:
        return []
    try:
        sentence_endings = re.compile(r'[.!?\n]')
        sentence_list = sentence_endings.split(text_string)
        return [sentence.strip() for sentence in sentence_list if sentence]
    except Exception as e:
        print(f"{e} while chunking {text_string}")
        return []

class Future:
    def __init__(self):
        self._event = threading.Event()
        self._result = None

    def set_result(self, result):
        self._result = result
        self._event.set()

    def result(self):
        self._event.wait()
        return self._result

class SimilaritySearchEngine:
    def __init__(self):
        self.last_message_id_future = Future()
        self.db_file = ".chatgpt-vectors.db"
        self.exists = os.path.isfile(fullpath(self.db_file))
        self.add_queue = queue.Queue()
        self.search_queue = queue.Queue()
        self.combo = queue.Queue()
        self.worker_thread = threading.Thread(target=self.worker)
        self.worker_thread.daemon = True
        self.worker_thread.start()

    def worker(self):
        self.vector_db = VectorDatabase(self.db_file)
        self.last_message_id_future.set_result(self.vector_db.last_message_id())
        while True:
            result = self.combo.get()
            search = None
            add = None
            try:
                # First check the search queue
                search = self.search_queue.get_nowait()
            except queue.Empty:
                try:
                    # If the search queue is empty, check the add queue
                    add = self.add_queue.get_nowait()
                except queue.Empty:
                    continue

            if add:
                message_id = None
                for (embedding, phrase) in add:
                    message_id = embedding.message_id
                    self.vector_db.add(embedding, phrase)
                if message_id is not None:
                    self.vector_db.commit_message(message_id)

            if search:
                search_results = self.vector_db.find_neighbors(*search)
                result.put(search_results)

    def add(self, text_string, chat_id, message_id):
        chunks = chunk_string(text_string)
        items = [(self.embedding(chunk, chat_id, message_id), chunk) for chunk in chunks]
        print(f"Index of chat={chat_id} message={message_id}: {len(items)} phrases")
        self.add_queue.put(items)
        self.combo.put([])

    def embedding(self, chunk, chat_id, message_id):
        v = compute_embedding(chunk)
        return Embedding(v, chat_id,  message_id)

    # Returns a list of message IDs
    def search(self, text_string):
        embedding = compute_embedding(text_string)
        distance = 0.7
        for attempt in range(4):
            self.search_queue.put([embedding, distance, 10])
            result = queue.Queue()
            self.combo.put(result)
            elist = result.get()
            if elist:
                return [(e.chat_id, e.message_id) for e in elist]
            distance *= 1.2

    def last_message_id(self):
        return self.last_message_id_future.result()





from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML
from src.highlight import SyntaxHighlighter
import datetime
import html
import os

def formattedTime(timestamp):
    dt = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
    dt = dt.replace(tzinfo=datetime.timezone.utc)
    dt = dt.astimezone()
    return dt.strftime("%b %d, %Y at %I:%M %p")

def print_message(timestamp, role, content, deleted, prefix=""):
    s = prefix + f"[{html.escape(formattedTime(timestamp))}] <i>{html.escape(role)}</i>:"
    if deleted:
        print_formatted_text(HTML("<strike>" + s + "</strike>"))
    else:
        print_formatted_text(HTML(s))
    if role == "assistant":
        sh = SyntaxHighlighter()
        sh.put(content)
        sh.eof()
    else:
        print(content, end='')


def setMark():
    if 'TERM' in os.environ and os.environ['TERM'].startswith('screen'):
        osc = "\033Ptmux;\033\033]"
        st = "\a\033\\"
    else:
        osc = "\033]"
        st = "\a"

    print(f"{osc}1337;SetMark{st}", end='')




import sys
import threading
import time

class Spinner:
    def __init__(self):
        self.i = 0
        self.running = True
        self.thread = threading.Thread(target=self._run)
        self.thread.start()
        self.wrote = False

    def stop(self):
        self.running = False
        self.thread.join()
        if self.wrote:
            sys.stdout.write(chr(8) + " " + chr(8))
            sys.stdout.flush()

    def _run(self):
        bs = ""
        spinner = ["|", "/", "-", "\\"]
        i = 0
        while self.running:
            sys.stdout.write(bs + spinner[i % 4])
            bs = chr(8)
            sys.stdout.flush()
            self.wrote = True
            time.sleep(0.1)
            i += 1

def spin(f):
    s = Spinner()
    result = f()
    s.stop()
    return result




import simplejson
import enum
import inspect
import openai
import sys
import threading
import time
import typing
from src.spin import spin

def get_json_type_name(value):
    if isinstance(value, str):
        return "string"
    elif isinstance(value, int):
        return "integer"
    elif isinstance(value, float):
        return "number"
    elif isinstance(value, bool):
        return "boolean"
    else:
        return "null"

def _json_schema(func):
    api_info = {}

    # Get function name
    api_info["name"] = func.__name__

    # Get function description from docstring
    docstring = inspect.getdoc(func)
    if docstring:
        api_info["description"] = docstring.split("\n\n")[0]

    # Get function parameters
    parameters = {}
    parameters["type"] = "object"

    properties = {}

    signature = inspect.signature(func)
    required = []
    for param_name, param in signature.parameters.items():
        param_info = {}

        # Get parameter type from type hints
        if typing.get_origin(param.annotation) is typing.Union and type(None) in typing.get_args(param.annotation):
            # Handle Optional case
            inner_type = typing.get_args(param.annotation)[0]
            if issubclass(inner_type, enum.Enum):
                param_info["type"] = get_json_type_name(inner_type.__members__[list(inner_type.__members__)[0]].value)
                param_info["enum"] = [member.value for member in inner_type]
            else:
                param_info["type"] = get_json_type_name(inner_type.__name__)
        elif issubclass(param.annotation, enum.Enum):
            param_info["type"] = get_json_type_name(param.annotation.__members__[list(param.annotation.__members__)[0]].value)
            param_info["enum"] = [member.value for member in param.annotation]
        else:
            param_info["type"] = get_json_type_name(param.annotation.__name__)

        # Get parameter description from docstring
        if docstring and param_name in docstring:
            param_info["description"] = docstring.split(param_name + ":")[1].split("\n")[0].strip()

        # Check if parameter is required
        if param.default == inspect.Parameter.empty:
            required.append(param_name)

        # Add parameter info to parameters dict
        properties[param_name] = param_info

    # Add parameters to api_info
    parameters["properties"] = properties
    parameters["required"] = required
    api_info["parameters"] = parameters

    return api_info

def create_chat_with_spinner(messages, temperature, functions, model):
    return create_chat(messages, temperature, functions, model, True)

def create_chat(messages, temperature, functions, model, spinner=False):
    def create_chat_model():
        args = {
            "model": model,
            "messages": messages,
            "stream": True,
            "temperature": temperature,
        }
        if functions:
            args['functions'] = list(map(lambda f: _json_schema(f), functions))
        return openai.ChatCompletion.create(**args)

    if not spinner:
        chats = []
        return create_chat_model()
    else:
        return spin(create_chat_model)


def suggest_name(chat_id, message):
    chat_completion_resp = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[
            {"role": "system", "content": "You assign names to conversations based on the first message. Respond with only a short, descriptive title for a conversation."},
            {"role": "user", "content": message}
        ],
        temperature=0,
        max_tokens=10
    )
    name = chat_completion_resp.choices[0].message.content
    return (chat_id, name)

def invoke(functions, name, args_str):
    try:
        args = simplejson.loads(args_str, strict=False)
    except Exception as e:
        return invoke_sloppy(functions, name, args_str)

    for func in functions:
        if func.__name__ == name:
            return func(**args)
    else:
        raise ValueError(f"Function '{name}' not found.")

def invoke_sloppy(functions, name, args_str):
    for func in functions:
        if func.__name__ == name and func.sloppy:
            args = [None] * len(inspect.signature(func).parameters)
            args[0] = args_str
            return func(*args)
    else:
        raise ValueError(f"Function '{name}' not found.")
